import logging
from collections.abc import Mapping, Sequence
from importlib.metadata import version
from typing import Any, Optional

import aiohttp
import aiohttp.web
from aiohttp.client import _RequestContextManager  # eventually will be public
from aiohttp.hdrs import AUTHORIZATION, USER_AGENT
from yarl import URL

from neuro_notifications_client.schema import SLUG_TO_SCHEMA

from .notifications import Notification

logger = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        url: Optional[URL],
        token: str,
        trace_configs: Sequence[aiohttp.TraceConfig] = (),
    ) -> None:
        if url is not None and not url:
            raise ValueError(
                "url argument should be http URL or None for secure-less configurations"
            )
        self._url = url
        self._headers: Mapping[str, str] = {
            AUTHORIZATION: f"Bearer {token}",
            USER_AGENT: f"NotificationsClient/{version(__package__)}",
        }
        self._trace_configs = list(trace_configs)
        self._client: Optional[aiohttp.ClientSession] = None

    async def init(self) -> None:
        if self._url is None:
            return
        if self._client is None:  # pragma: no branch
            self._client = aiohttp.ClientSession(
                raise_for_status=True,
                headers=self._generate_headers(),
                trace_configs=self._trace_configs,
            )

    async def __aenter__(self) -> "Client":
        await self.init()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    async def close(self) -> None:
        if self._client is not None:
            client = self._client
            self._client = None
            await client.close()

    def _generate_headers(self) -> Mapping[str, str]:
        return self._headers

    def _make_url(self, path: str) -> URL:
        assert self._url
        if path.startswith("/"):
            path = path[1:]
        return self._url / path

    def _request(
        self, method: str, path: str, *args: Any, **kwargs: Any
    ) -> _RequestContextManager:
        # Not a coroutine actually but coroutine builder
        # later we'll drop the trick most likely
        url = self._make_url(path)
        return self._client.request(method, url, *args, **kwargs)  # type: ignore

    async def ping(self, timeout_seconds: float = 10) -> None:
        if self._url is None:
            return
        timeout = aiohttp.ClientTimeout(total=timeout_seconds)
        async with self._request("GET", "/api/v1/ping", timeout=timeout) as resp:
            txt = await resp.text()
            assert txt == "Pong"

    async def secured_ping(self, timeout_seconds: float = 10) -> None:
        if self._url is None:
            return
        timeout = aiohttp.ClientTimeout(total=timeout_seconds)
        async with self._request(
            "GET", "/api/v1/secured-ping", timeout=timeout
        ) as resp:
            txt = await resp.text()
            assert txt == "Pong"

    async def notify(self, notification: Notification) -> None:
        if self._url is None:
            return
        slug = notification.slug()
        schema_cls = SLUG_TO_SCHEMA.get(slug)
        if schema_cls is None:
            raise ValueError(f"Notification {notification} is not supported")
        payload = schema_cls().dump(notification)
        async with self._request(
            "POST", f"/api/v1/notifications/{slug}", json=payload
        ) as resp:
            if resp.status != aiohttp.web.HTTPCreated.status_code:
                raise RuntimeError("Server return %s instead HTTPCreated", resp.status)



from __future__ import annotations

import abc
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Optional


class Notification(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def slug(cls) -> str:
        pass


@dataclass
class Welcome(Notification):
    user_id: str
    email: str

    @classmethod
    def slug(cls) -> str:
        return "welcome"


@dataclass
class Invite(Notification):
    org_name: str
    email: str
    console_url: str

    @classmethod
    def slug(cls) -> str:
        return "invite"


@dataclass
class JobNotification(Notification, abc.ABC):
    job_id: str


@dataclass
class JobCannotStartLackResources(JobNotification):
    @classmethod
    def slug(cls) -> str:
        return "job-cannot-start-lack-resources"


@dataclass
class JobTransition(JobNotification):
    job_id: str
    status: str
    transition_time: datetime
    reason: Optional[str] = None
    description: Optional[str] = None
    exit_code: Optional[int] = None
    prev_status: Optional[str] = None
    prev_transition_time: Optional[datetime] = None

    @classmethod
    def slug(cls) -> str:
        return "job-transition"


class QuotaResourceType(str, Enum):
    NON_GPU = "non_gpu"
    GPU = "gpu"


@dataclass
class JobCannotStartQuotaReached(Notification):
    user_id: str
    resource: QuotaResourceType
    quota: float
    cluster_name: str

    @classmethod
    def slug(cls) -> str:
        return "job-cannot-start-quota-reached"


@dataclass
class QuotaWillBeReachedSoon(Notification):
    user_id: str
    resource: QuotaResourceType
    used: float
    quota: float
    cluster_name: str

    @classmethod
    def slug(cls) -> str:
        return "quota-will-be-reached-soon"


@dataclass
class JobCannotStartNoCredits(Notification):
    user_id: str
    cluster_name: Optional[str] = None

    @classmethod
    def slug(cls) -> str:
        return "job-cannot-start-no-credits"


@dataclass
class CreditsWillRunOutSoon(Notification):
    user_id: str
    cluster_name: str
    credits: Decimal

    @classmethod
    def slug(cls) -> str:
        return "credits-will-run-out-soon"


@dataclass
class AlertManagerNotification(Notification):
    class Status(str, Enum):
        RESOLVED = "resolved"
        FIRING = "firing"

    @dataclass
    class Alert:
        status: AlertManagerNotification.Status
        labels: dict[str, str]
        annotations: dict[str, str]

    version: str
    group_key: str
    status: Status
    group_labels: dict[str, str]
    common_labels: dict[str, str]
    common_annotations: dict[str, str]
    alerts: list[Alert]

    @classmethod
    def slug(cls) -> str:
        return "alert-manager-notification"



from datetime import datetime

import pytest

from neuro_notifications_client import (
    Client,
    JobCannotStartLackResources,
    JobCannotStartQuotaReached,
    JobTransition,
    QuotaWillBeReachedSoon,
)
from neuro_notifications_client.notifications import (
    AlertManagerNotification,
    JobNotification,
    QuotaResourceType,
    Welcome,
    Invite,
)

# TODO: make this tests more meaningful:
# at the moment they just test that no exception was raised


@pytest.mark.parametrize(
    "notification",
    [
        JobTransition(
            "job-new-transition",
            status="PENDING",
            transition_time=datetime(year=2019, month=6, day=7),
        ),
        JobTransition(
            "job-started",
            status="RUNNING",
            transition_time=datetime(year=2019, month=6, day=8),
            prev_status="PENDING",
            prev_transition_time=datetime(year=2019, month=6, day=7),
        ),
        JobTransition(
            "job-failed-exit",
            status="FAILED",
            transition_time=datetime(year=2019, month=6, day=9),
            prev_status="RUNNING",
            prev_transition_time=datetime(year=2019, month=6, day=8),
            reason="Exited with code -10",
            exit_code=-10,
        ),
        JobTransition(
            "job-failed-image",
            status="FAILED",
            transition_time=datetime(year=2019, month=6, day=10),
            prev_status="PENDING",
            prev_transition_time=datetime(year=2019, month=6, day=9),
            reason="ImagePullBackOff",
            description="Cannot pull image some:tag",
        ),
        JobTransition(
            "job-finished",
            status="SUCCEEDED",
            transition_time=datetime(year=2019, month=6, day=11),
            prev_status="RUNNING",
            prev_transition_time=datetime(year=2019, month=6, day=10),
            exit_code=0,
        ),
    ],
)
async def test_job_transition_notifications(
    client: Client,
    notification: JobNotification,
) -> None:
    # Should not raise anything
    await client.notify(notification)
    await client.close()


@pytest.mark.parametrize(
    "notification", [JobCannotStartLackResources("job-lack-resources")]
)
async def test_job_notifications(
    client: Client,
    notification: JobNotification,
) -> None:
    # Should not raise anything
    await client.notify(notification)
    await client.close()


@pytest.mark.parametrize(
    "notification",
    [
        JobCannotStartQuotaReached(
            user_id="job-quota-reached",
            resource=QuotaResourceType.NON_GPU,
            quota=2.1,
            cluster_name="test",
        )
    ],
)
async def test_user_notifications(
    client: Client,
    notification: JobNotification,
) -> None:
    # Should not raise anything
    await client.notify(notification)
    await client.close()


@pytest.mark.parametrize(
    "notification",
    [
        QuotaWillBeReachedSoon(
            "job-quota-reached",
            resource=QuotaResourceType.NON_GPU,
            used=1,
            quota=2,
            cluster_name="default",
        ),
        QuotaWillBeReachedSoon(
            "job-quota-reached",
            resource=QuotaResourceType.GPU,
            used=2,
            quota=3,
            cluster_name="default",
        ),
    ],
)
async def test_quota_will_be_reached_soon_notifications(
    client: Client,
    notification: QuotaWillBeReachedSoon,
) -> None:
    # Should not raise anything
    await client.notify(notification)
    await client.close()


async def test_welcome_notification(client: Client) -> None:
    # Should not raise anything
    await client.notify(Welcome(user_id="bob", email="bob@neu.ro"))
    await client.close()


async def test_invite_notification(client: Client) -> None:
    # Should not raise anything
    await client.notify(
        Invite(org_name="test", email="bob@neu.ro", console_url="test-url")
    )
    await client.close()


async def test_alert_manager_notification(client: Client) -> None:
    # Should not raise anything
    await client.notify(
        AlertManagerNotification(
            version="4",
            group_key='{}:{alertname="TraefikDown"}',
            status=AlertManagerNotification.Status.FIRING,
            group_labels={"alertname": "TraefikDown"},
            common_labels={
                "alertname": "TraefikDown",
                "cluster": "default",
                "context": "traefik",
                "severity": "critical",
            },
            common_annotations={"summary": "Traefik Down"},
            alerts=[
                AlertManagerNotification.Alert(
                    status=AlertManagerNotification.Status.FIRING,
                    labels={
                        "alertname": "TraefikDown",
                        "cluster": "default",
                        "severity": "critical",
                    },
                    annotations={"summary": "Traefik Down"},
                ),
            ],
        )
    )
    await client.close()


async def test_unknown_notifications(
    client: Client,
    unknown_notification: JobNotification,
) -> None:
    with pytest.raises(ValueError):
        await client.notify(unknown_notification)
        await client.close()



import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable

import aiohttp.web
import pytest
from aiohttp.hdrs import AUTHORIZATION
from aiohttp.test_utils import TestServer as AioHTTPTestServer
from aiohttp.web import Application

from neuro_notifications_client import Client
from neuro_notifications_client.notifications import Notification


@pytest.fixture
async def api_server(
    aiohttp_server: Callable[[Application], Awaitable[AioHTTPTestServer]], token: str
) -> AsyncIterator[AioHTTPTestServer]:
    async def _post_notification(request: aiohttp.web.Request) -> aiohttp.web.Response:
        raise aiohttp.web.HTTPCreated()

    async def _unknown_notification(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        raise aiohttp.web.HTTPNotFound()

    async def _strange_notification(
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        raise aiohttp.web.HTTPOk()

    async def _ping(request: aiohttp.web.Request) -> aiohttp.web.Response:
        await asyncio.sleep(2)
        return aiohttp.web.Response(text="Pong")

    async def _secured_ping(request: aiohttp.web.Request) -> aiohttp.web.Response:
        auth_header = request.headers.get(AUTHORIZATION)
        if auth_header != f"Bearer {token}":
            raise aiohttp.web.HTTPForbidden
        await asyncio.sleep(2)
        return aiohttp.web.Response(text="Pong")

    app = aiohttp.web.Application()
    app.add_routes([aiohttp.web.get("/api/v1/ping", _ping)])
    app.add_routes([aiohttp.web.get("/api/v1/secured-ping", _secured_ping)])

    app.add_routes(
        [aiohttp.web.post("/api/v1/notifications/unknown", _unknown_notification)]
    )
    app.add_routes(
        [aiohttp.web.post("/api/v1/notifications/{type}", _post_notification)]
    )
    server = await aiohttp_server(app)
    yield server


@pytest.fixture
def token() -> str:
    return "secured-token"


@pytest.fixture
async def client(api_server: AioHTTPTestServer, token: str) -> AsyncIterator[Client]:

    async with Client(api_server.make_url(""), token=token) as client:
        yield client


@pytest.fixture
def unknown_notification() -> Notification:
    class UnknownNotification(Notification):
        @classmethod
        def slug(cls) -> str:
            return "unknown"

    return UnknownNotification()



from __future__ import annotations

from datetime import timezone
from enum import Enum
from typing import Any, Optional

from marshmallow import EXCLUDE, Schema, fields, post_load, validate

from .notifications import (
    AlertManagerNotification,
    CreditsWillRunOutSoon,
    JobCannotStartLackResources,
    JobCannotStartNoCredits,
    JobCannotStartQuotaReached,
    JobTransition,
    QuotaResourceType,
    QuotaWillBeReachedSoon,
    Welcome,
    Invite,
)


class StringEnum(fields.String):
    def __init__(self, enum: type[Enum], *args: Any, **kwargs: Any) -> None:
        super().__init__(
            *args, validate=validate.OneOf([item.value for item in enum]), **kwargs
        )
        self.enum = enum

    def _deserialize(self, *args: Any, **kwargs: Any) -> Enum:
        res: str = super()._deserialize(*args, **kwargs)
        return self.enum(res)

    def _serialize(
        self, value: Optional[Enum], *args: Any, **kwargs: Any
    ) -> Optional[str]:
        if value is None:
            return None
        return super()._serialize(value.value, *args, **kwargs)


class JobCannotStartLackResourcesSchema(Schema):
    job_id = fields.String(required=True)

    @post_load
    def make_notification(
        self, data: Any, **kwargs: Any
    ) -> JobCannotStartLackResources:
        return JobCannotStartLackResources(**data)


class JobTransitionSchema(Schema):
    job_id = fields.String(required=True)
    status = fields.String(required=True)
    transition_time = fields.AwareDateTime(required=True, default_timezone=timezone.utc)
    reason = fields.String(required=False, allow_none=True)
    description = fields.String(required=False, allow_none=True)
    exit_code = fields.Integer(required=False, allow_none=True)
    prev_status = fields.String(required=False, allow_none=True)
    prev_transition_time = fields.AwareDateTime(
        required=False, allow_none=True, default_timezone=timezone.utc
    )

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> JobTransition:
        return JobTransition(**data)


class JobCannotStartNoCreditsSchema(Schema):
    user_id = fields.String(required=True)
    cluster_name = fields.String(required=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> JobCannotStartNoCredits:
        return JobCannotStartNoCredits(**data)


class CreditsWillRunOutSoonSchema(Schema):
    user_id = fields.String(required=True)
    cluster_name = fields.String(required=True)
    credits = fields.Decimal(required=True, as_string=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> CreditsWillRunOutSoon:
        return CreditsWillRunOutSoon(**data)


class JobCannotStartQuotaReachedSchema(Schema):
    user_id = fields.String(required=True)
    cluster_name = fields.String(required=True)
    resource = StringEnum(enum=QuotaResourceType, required=True)
    quota = fields.Float(required=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> JobCannotStartQuotaReached:
        return JobCannotStartQuotaReached(**data)


class QuotaWillBeReachedSoonSchema(Schema):
    user_id = fields.String(required=True)
    cluster_name = fields.String(required=True)
    resource = StringEnum(enum=QuotaResourceType, required=True)
    used = fields.Float(required=True)
    quota = fields.Float(required=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> QuotaWillBeReachedSoon:
        data["resource"] = QuotaResourceType(data["resource"])
        return QuotaWillBeReachedSoon(**data)


class WelcomeSchema(Schema):
    user_id = fields.String(required=True)
    email = fields.Email(required=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> Welcome:
        return Welcome(**data)


class InviteSchema(Schema):
    org_name = fields.String(required=True)
    email = fields.Email(required=True)
    console_url = fields.String(required=True)

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> Invite:
        return Invite(**data)


class AlertManagerNotificationSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    class AlertSchema(Schema):
        class Meta:
            unknown = EXCLUDE

        status = fields.Enum(
            AlertManagerNotification.Status, by_value=True, required=True
        )
        labels = fields.Dict(keys=fields.String(), values=fields.String())
        annotations = fields.Dict(keys=fields.String(), values=fields.String())

        @post_load
        def make_alert(
            self, data: Any, **kwargs: Any
        ) -> AlertManagerNotification.Alert:
            return AlertManagerNotification.Alert(**data)

    version = fields.String(validate=validate.Equal("4"), required=True)
    group_key = fields.String(data_key="groupKey")
    status = fields.Enum(AlertManagerNotification.Status, by_value=True, required=True)
    group_labels = fields.Dict(
        keys=fields.String(), values=fields.String(), data_key="groupLabels"
    )
    common_labels = fields.Dict(
        keys=fields.String(), values=fields.String(), data_key="commonLabels"
    )
    common_annotations = fields.Dict(
        keys=fields.String(), values=fields.String(), data_key="commonAnnotations"
    )
    alerts = fields.List(fields.Nested(AlertSchema), validate=validate.Length(min=1))

    @post_load
    def make_notification(self, data: Any, **kwargs: Any) -> AlertManagerNotification:
        return AlertManagerNotification(**data)


SLUG_TO_SCHEMA = {
    Invite.slug(): InviteSchema,
    Welcome.slug(): WelcomeSchema,
    JobCannotStartLackResources.slug(): JobCannotStartLackResourcesSchema,
    JobTransition.slug(): JobTransitionSchema,
    JobCannotStartNoCredits.slug(): JobCannotStartNoCreditsSchema,
    CreditsWillRunOutSoon.slug(): CreditsWillRunOutSoonSchema,
    QuotaWillBeReachedSoon.slug(): QuotaWillBeReachedSoonSchema,
    JobCannotStartQuotaReached.slug(): JobCannotStartQuotaReachedSchema,
    AlertManagerNotification.slug(): AlertManagerNotificationSchema,
}



from .client import Client
from .notifications import (
    AlertManagerNotification,
    CreditsWillRunOutSoon,
    JobCannotStartLackResources,
    JobCannotStartNoCredits,
    JobCannotStartQuotaReached,
    JobTransition,
    QuotaResourceType,
    QuotaWillBeReachedSoon,
    Welcome,
    Invite,
)

__all__ = [
    "Client",
    "JobCannotStartLackResources",
    "JobCannotStartQuotaReached",
    "JobCannotStartNoCredits",
    "JobTransition",
    "QuotaWillBeReachedSoon",
    "QuotaResourceType",
    "CreditsWillRunOutSoon",
    "Welcome",
    "Invite",
    "AlertManagerNotification",
]



from asyncio import TimeoutError

import pytest
from aiohttp.client_exceptions import ClientResponseError

from neuro_notifications_client import Client


async def test_ping(client: Client) -> None:
    await client.ping()  # should not raise exception


async def test_ping_with_timeout(client: Client) -> None:
    with pytest.raises(TimeoutError):
        await client.ping(0.1)


async def test_secured_ping(client: Client) -> None:
    await client.secured_ping()  # should not raise exception
    async with Client(client._url, "no-access-token") as no_access_client:
        with pytest.raises(ClientResponseError):
            await no_access_client.secured_ping()


async def test_secured_ping_with_timeout(client: Client) -> None:
    with pytest.raises(TimeoutError):
        await client.secured_ping(0.1)
